/* Copyright (C) 2004 Hiroyuki OYAMA

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*
  storage.c - default storage function.
  see also 'HACK' file.
*/

#include "httpd.h"
#include "http_main.h"
#include "http_log.h"

#include "trackback.h"
#include "storage.h"
#include <stdio.h>
#include <fcntl.h>


static const char *get_storage_path(pool *p, trackback_conf *conf, const char *id)
{
    const char *path;

    if (conf->storage == NULL) {
        path = ap_pstrcat(p, "/tmp/", id, NULL);
    }
    else {
        path = ap_pstrcat(p, conf->storage, "/", id, NULL);
    }
    return path;
}

static char *each_word(pool *p, char **src, char stop)
{
    char *pos = strchr(*src, stop);
    char *res;

    if (!pos) {
        res = ap_pstrdup(p, *src);
        *src += strlen(*src);
        return res;
    }

    res = ap_pstrndup(p, *src, pos - *src);
    ++pos;
    *src = pos;
    return res;
}


static int lock_fd(int fd, int type)
{
    struct flock lock;

    lock.l_type = type;
    lock.l_start = 0;
    lock.l_whence = SEEK_SET;
    lock.l_len = 0;
    return fcntl(fd, F_SETLKW, &lock);
}

static int unlock_fd(int fd)
{
    struct flock lock;

    lock.l_type = F_UNLCK;
    lock.l_start = 0;
    lock.l_whence = SEEK_SET;
    lock.l_len = 0;
    return fcntl(fd, F_SETLK, &lock);
}


int mtb_store_message(request_rec *r, mtb_trackback_message_rec *msg)
{
    trackback_conf *conf;
    char *comment;
    const char *path;
    int fd;

    conf = ap_get_module_config(r->per_dir_config, &trackback_module);
    comment = ap_psprintf(r->pool, "%d %s %s %s %s\n",
                         (unsigned int)msg->date,
                         ap_escape_uri(r->pool, msg->title),
                         ap_escape_uri(r->pool, msg->excerpt),
                         ap_escape_uri(r->pool, msg->blog_name),
                         ap_escape_uri(r->pool, msg->url));

    path = get_storage_path(r->pool, conf, msg->id);
    ap_hard_timeout("append trackback", r);
    if ((fd = ap_popenf(r->pool, path, O_WRONLY|O_APPEND, 0644)) < 0) {
        ap_kill_timeout(r);
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                      "cannot append trackback db: %s", path);
        return FORBIDDEN;
    }
    lock_fd(fd, F_WRLCK);
    write(fd, comment, strlen(comment));
    unlock_fd(fd);
    ap_pclosef(r->pool, fd);
    ap_kill_timeout(r);

    return OK;
}

mtb_trackback_message_rec *mtb_load_message(request_rec *r, const char *id)
{
    trackback_conf *conf;
    mtb_trackback_message_rec *msg, *head, *newmsg;
    const char *path;
    FILE *fp;
    unsigned char buff[HUGE_STRING_LEN];

    conf = ap_get_module_config(r->per_dir_config, &trackback_module);
    path = get_storage_path(r->pool, conf, id);
    if ((fp = ap_pfopen(r->pool, (char *)path, "r")) == NULL) {
        return NULL;
    }

    msg = NULL;
    head = NULL;
    lock_fd(fileno(fp), F_RDLCK);
    while (fgets(buff, sizeof(buff), fp) != NULL) {
        char *date, *title, *excerpt, *blog_name, *url, *pos, *tail;
        time_t postdate;

        pos = buff;
        date      = each_word(r->pool, &pos, ' ');
        title     = each_word(r->pool, &pos, ' ');
        excerpt   = each_word(r->pool, &pos, ' ');
        blog_name = each_word(r->pool, &pos, ' ');
        url       = each_word(r->pool, &pos, ' ');
        if (date == NULL || title == NULL || excerpt == NULL
            || blog_name == NULL || url == NULL) {
            continue;
        }
        postdate = (time_t)strtol(date, (char **)NULL, 10);
        if (postdate <= 0)
            continue;
        if ((tail = strchr(url, '\n')) != NULL)
            *tail = '\0';
        ap_unescape_url(title);
        ap_unescape_url(excerpt);
        ap_unescape_url(blog_name);
        ap_unescape_url(url);
        
        newmsg = ap_palloc(r->pool, sizeof(mtb_trackback_message_rec));
        newmsg->id = ap_pstrdup(r->pool, id);
        newmsg->date      = postdate;
        newmsg->title     = title;
        newmsg->excerpt   = excerpt;
        newmsg->blog_name = blog_name;
        newmsg->url       = url;
        newmsg->next      = NULL;
        if (msg != NULL) {
            msg->next = newmsg;
        }
        else {
            head = newmsg;
        }
        msg = newmsg;
    }
    unlock_fd(fileno(fp));
    ap_pfclose(r->pool, fp);

    return head;
}


